/*
	Jakub Lewczyski, 3T2

	Opis: Obiekt copy-on-write proxy, tak jak zwyky proxy, poredniczy miedzy klientem a rzeczywistym obiektem.
	Jego zadaniem jest wykonanie kopiowania obiektu dopiero w momencie wprowadzania do niego zmian.
	Kada kolejna kopia obiektu nie zajmuje dodatkowej pamici. Gdy mamy do czynienia z obiektami o znacznym rozmiarze
	jest to dua oszczdno. Wad tego rozwizania jest duszy czas modyfikacji obiektu.

	Dziaanie: Klasa BigClassProxy przechowuje wskaniki na: obiekt BigClass i licznik odwoa do tego obiektu.
	Kiedy obiekt ma by zmodyfikowany proxy sprawdza licznik odwoa. Jeli do danego obiektu odowuje si tylko
	jedno proxy - obiekt jest modyfikowany. Jeli do obiektu odwouje si wicej proxy - proxy modyfikujce obiekt
	zmniejsza dotychczasowy licznik odwoa, tworzy nowy obiekt na podstawie dotychczasowego, tworzy nowy licznik
	i na sam koniec wykonuje dane zmiany. W trakcie przypisywania proxy dziaa analogicznie, tyle e sprawdza czy
	zmniejszany licznik odwoa nie wynosi 0. Jeli tak jest obiekt, do ktrego nie odwouje si adne proxy jest zwalniany.
*/

#include <iostream>

/*Abstrakcyjna klasa bazowa dla klas BigClass i Big classProxy. Zapewnia im wsplny interfejs.*/
class BigClassInterface {
public:
	virtual ~BigClassInterface() {}
	virtual void setValue(int value) = 0;
	virtual int getValue() const = 0;
protected:
	BigClassInterface() {}
};

/*Klasa, ktrej kopiowanie nastpuje przy zmianie jej parametrw. Dziedziczy po abstrakcyjnej klasie interfejsu.*/
class BigClass: public BigClassInterface {
public:
	BigClass(int value): value_(value) { std::cout << "Konstruktor duzego obiektu" << std::endl; }	//konstruktor duej klasy
	BigClass(const BigClass& source): value_(source.value_) { std::cout << "Konstruktor duzego obiektu" << std::endl; }	//konstruktor kopiujcy
	void setValue(int value) { value_ = value; }	//funkcja zapisujca przechowywane dane (w tym wypadku int)
	int getValue() const { return value_; }				//funkcja zwracajca przechowywane dane (w tym wypadku int)
	~BigClass() { std::cout << "Destruktor proxy." << std::endl; }	//destruktor
private:
	int value_;	//przechowywane dane
};

/*Proxy klasy BigClass. Dziedziczy po abstrakcyjnej klasie interfejsu.*/
class BigClassProxy: public BigClassInterface {
public:
	BigClassProxy(int value);	//konstruktor
	BigClassProxy(const BigClassProxy& source); //konstruktor kopiujcy

	void setValue(int value) { real()->setValue(value); }	//funkcja kopiujca obiekt i wprowadzajca do niego zmiany
	int getValue() const { return realPtr_->getValue(); }	//funkcja zwracajca dane przechowywane w obiekcie

	BigClassProxy& operator=(const BigClassProxy& source);	//operator przypisania

	~BigClassProxy();	//destruktor proxy
private:
	BigClass* realPtr_;	//wskanik na obiekt
	int* counter_;		//wskanik na licznik odniesie do obiektu
	/*Funkcja pomocnicza obsugujca ewentualne kopiowanie obiektu i zwracajca do niego wskanik*/
	BigClass* real() {
		//jeli wicej ni jedno proxy odwouje si do obiektu...
		if(*counter_ > 1) {
			//zmniejsz licznik odwoa
			(*counter_)--;

			//stworzenie nowego licznika odniesie i obiekt BigClass na podstawie starego obiektu
			counter_ = new int(1);
			realPtr_ = new BigClass(*realPtr_);
		}
		//jeli tylko jedno proxy korzysta z obiektu nie trzeba tworzy nowego obiektu; wystarczy zmodyfikowa obiekt
		//zwr aktualny wskanik
		return realPtr_;
	}
};

/*Konstruktor proxy*/
BigClassProxy::BigClassProxy(int value){
	//komunikat informacyjny
	std::cout << "Konstruktor proxy." << std::endl;

	//stworzenie nowego obiektu i licznika odniesie
	realPtr_ = new BigClass(value);
	counter_ = new int(1);
}

/*Konstruktor kopiujcy proxy*/
BigClassProxy::BigClassProxy(const BigClassProxy& source){
	//komunikat informacyjny
	std::cout << "Konstruktor kopiujcy proxy." << std::endl;

	//przypisanie adresu obiektu i licznika odniesie
	realPtr_ = source.realPtr_;
	counter_ = source.counter_;

	//zwikszenie licznika odniesie
	(*counter_)++;
}

/*Operator przypisania proxy*/
BigClassProxy& BigClassProxy::operator=(const BigClassProxy& source) {
	//tymczsowe zmienne przechowujce wskaniki do obiektu i licznika
	int* old_counter = counter_;
	BigClass* old_real = realPtr_;

	//przypisanie nowego obiektu i wskanika do wskanikw
	counter_ = source.counter_;
	realPtr_ = source.realPtr_;

	//zmniejszenie licznika odniesie do starego obiektu
	(*old_counter)--;
	//zwikszenie licznika odniesie do nowego obiektu
	(*counter_)++;

	//jeli nikt nie uywa starego obiektu zwalnia pami
	if(*old_counter == 0) {
		delete old_counter;
		delete old_real;
	}

	//zwraca referencj do siebie
	return *this;
}

/*Destrukto proxy*/
BigClassProxy::~BigClassProxy(){
	//komunikat informacyjny
	std::cout << "Destruktor proxy." << std::endl;

	//zmiejszenie licznika odniesie
	(*counter_)--;

	//jeli nikt nie uywa obiektu zwalnia pami
	if(*counter_ == 0) {
		delete realPtr_;
		delete counter_;
	}
}

int main()
{
	std::cout << "Poczatek programu" << std::endl;
	std::cout << "Stworzenie proxy 'a' o wartoci 10" << std::endl;
	BigClassProxy a(10);
	std::cout << "Stworzenie proxy 'b' na podstawie 'a'" << std::endl;
	BigClassProxy b(a);
	std::cout << "Stworzenie proxy 'c' na podstawie 'a'" << std::endl;
	BigClassProxy c(a);
	std::cout << "Stworzenie proxy 'd' na podstawie 'a'" << std::endl;
	BigClassProxy d(a);

	std::cout << "Wypisanie wartoci:" << std::endl;
	std::cout << "'a': " << a.getValue() << std::endl;
	std::cout << "'b': " << b.getValue() << std::endl;
	std::cout << "'c': " << c.getValue() << std::endl;
	std::cout << "'d': " << d.getValue() << std::endl;

	std::cout << "Ustawienie wartoci 'b' na 50" << std::endl;
	b.setValue(50);
	std::cout << "Przypisanie 'b' do 'c'" << std::endl;
	c = b;
	std::cout << "Ustawienie wartoci 'd' na 100" << std::endl;
	d.setValue(100);
	std::cout << "Przypisanie 'd' do 'a'" << std::endl;
	a = d;	

	std::cout << "Wypisanie wartoci:" << std::endl;
	std::cout << "'a': " << a.getValue() << std::endl;
	std::cout << "'b': " << b.getValue() << std::endl;
	std::cout << "'c': " << c.getValue() << std::endl;
	std::cout << "'d': " << d.getValue() << std::endl;
	return 0;
}

